#!/usr/bin/python

## Code for converting a circuit into a SAT instance
## which produces an n x m lattice, by unfolding the
## quantified variables.
import sys
import qbf
from constraints import sorter, cardnet, atmost_one, reify
from formula import ExpTable, BoolExp
from itertools import chain
import time

#start_time_prep2=time.time()
target_idx = 0

ncols = int(sys.argv[-1])
nrows = int(sys.argv[-2])

problem = qbf.QBF()

## Assume input is in pla format.
invars = set()
terms = []
for line in sys.stdin.readlines():
  [sin, sout] = line.strip().split()
  ## The given minterm has f(X) = 1
  if sout[target_idx] == '1':
    term = []
    for ti in xrange(len(sin)):
      vi = ti+1
      if sin[ti] == '1':
        term.append(vi)
        invars |= { vi }
      elif sin[ti] == '0':
        term.append(-vi)
        invars |= { vi }

    terms.append(term)
    
# print terms

## Convert the set of input variables
## to a list.
invars = list(invars)

## Set up the expression table
table = ExpTable()
def toExp(v):
  return table.leaf(v)

## Construct the quantified variables
## Ordering [F, T, x0, ~x0, x1, ~x1, ...]
latvars = { (i, j) : problem.existFD(2 + 2*len(invars))
              for i in xrange(nrows) for j in xrange(ncols) }
## qvars = { v : problem.univ() for v in invars }

#for (i, j) in latvars:
#  print >> sys.stderr, (i, j), ":", latvars[(i, j)]

## Collect the path variables adjacent to a square.
def pos_adjvars(vs, i, j):
  if i == 0:
    return [ vs[(i+1, j)] ]
  if i == nrows-1:
    return [ vs[(i-1, j)] ]
  return [ vs[elt] for elt in [(i-1, j), (i+1, j), (i, j-1), (i, j+1)] if elt in vs ]

def neg_adjvars(vs, i, j):
  if j == 0:
    return [ vs[(ip, j+1)] for ip in [i-1, i, i+1] if (ip, j+1) in vs ]
  if j == ncols-1:
    return [ vs[(ip, j-1)] for ip in [i-1, i, i+1] if (ip, j-1) in vs ]
  return [ vs[elt] for elt in 
            [(i-1, j-1), (i-1, j), (i-1, j+1), (i, j-1), (i, j+1), (i+1, j-1), (i+1, j), (i+1,j+1)]
            if elt in vs ]

##pos_nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
##neg_nvars = { (i, j) : problem.exist() for i in xrange(nrows) for j in xrange(ncols) }

## Build the positive instances for a given assignment
def build_pos(asg):
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    ## ~switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    assert len(lvar) == 2*len(asg) + 2

    problem.addClause([~lvar[0], ~nv])
    #print >>sys.stderr,"basla", [~lvar[0], ~nv]
    for vi in xrange(len(asg)):
      if asg[vi]:
        problem.addClause([~lvar[2*vi + 3], ~nv])
      else:
        problem.addClause([~lvar[2*vi + 2], ~nv])

  if nrows == 1:
    problem.addClause(nvars.values())
  elif ncols == 1:
    for val in nvars.values():
      problem.addClause([val])
  else:
    for (i, j) in nvars.keys():
      nv = nvars[(i, j)]
      adjs = pos_adjvars(nvars, i, j)
      assert len(adjs) in [1, 3, 4]
      
      if len(adjs) == 1:
        ## Terminal edge.
        problem.addClause([adjs[0],~nv])
      else:
        assert len(adjs) >= 3
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        ## nv -> sum(adjs) = 2
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])

    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == nrows-1 ], problem.ttt())
    
## And the negative paths
def build_neg(asg):
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    ## switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    assert len(lvar) == 2*len(asg) + 2

    problem.addClause([~lvar[1], ~nv]) ## [[ lij = T ]] => ~neg[i,j]
    #print >>sys.stderr,"baslama", [~lvar[1], ~nv]
    for vi in xrange(len(asg)):
      if asg[vi]:
        problem.addClause([~lvar[2*vi + 2], ~nv])
        #print >>sys.stderr,"baslama", [~lvar[2*vi + 2], ~nv]
      else:
        problem.addClause([~lvar[2*vi + 3], ~nv])
        #print >>sys.stderr,"baslama", [~lvar[2*vi + 3], ~nv]

  if ncols == 1:
    problem.addClause(nvars.values())
  elif nrows == 1:
    for val in nvars.values():
      problem.addClause([val])
  else:
    for (i, j) in nvars.keys():
      ## Connect up the path variables.
      nv = nvars[(i, j)]
      adjs = neg_adjvars(nvars, i, j)
      #print >> sys.stderr, i, j, adjs
      assert (len(adjs) in [2, 3, 5, 8])
      
      
      if len(adjs) <= 2:
        ## Terminal edge.
        assert j in [0, ncols-1]
        #adjv = problem.exist()
        #atmost_one(problem, adjs, adjv)
        #problem.addClause([adjv, ~nv])
        problem.addClause([adjs[0],adjs[1], ~nv])
        problem.addClause([~adjs[0],~adjs[1], ~nv])
        problem.addClause([adjs[0],adjs[1], ~nv])
      elif len(adjs) == 3:
        ## Terminal edge.
        assert j in [0, ncols-1]
        #adjv = problem.exist()
        #atmost_one(problem, adjs, adjv)
        #problem.addClause([adjv, ~nv])
        
        problem.addClause([adjs[0],adjs[1],adjs[2], ~nv])
        #adjv = problem.exist()
        #problem.addClause([adjs[0],adjs[1],adjv])
        #problem.addClause([~adjv,adjs[2], ~nv])
        
        problem.addClause([~adjs[0],~adjs[1],~adjs[2], ~nv])
        #adjv = problem.exist()
        #problem.addClause([~adjs[0],~adjs[1],adjv])
        #problem.addClause([~adjv,~adjs[2], ~nv])
        problem.addClause([~adjs[1],~adjs[2], ~nv])
        problem.addClause([~adjs[0],~adjs[2], ~nv])
        problem.addClause([~adjs[0],~adjs[1], ~nv])
      else:
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])
	
    #problem.addClause([nvars[(0,2)], nvars[(1,2)], nvars[(2,2)], nvars[(3,2)]])
	
	## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == ncols-1 ], problem.ttt())

def all_assigns(k):
  if k <= 0:
    return [ [] ]
  else:
    rest = all_assigns(k-1)
    return map(lambda asg: [False] + asg, rest) + map(lambda asg: [True] + asg, rest)

def eval_assign(asg):
  def eval_elt(elt, asg):
    if elt > 0:
#      print elt
      return asg[elt-1]
    else:
      return not asg[-elt-1]
  def eval_term(term, asg):
    return reduce(lambda x, y: x and eval_elt(y, asg), term, True)

  return reduce(lambda x, y: x or eval_term(y, asg), terms, False) 

for asg in all_assigns(len(invars)):
#  print >> sys.stderr, asg
  if eval_assign(asg):
#    print >> sys.stderr, "T:", asg
    build_pos(asg)
  else:
#    print >> sys.stderr, "F:", asg
    build_neg(asg)


problem.to_dimacs()
#end_time3=time.time()
#print >>sys.stderr, end_time3-start_time_prep2, " :time prepare input to SAT for optimal"
