## Constraints.py 
## ==============
## Assorted utility constraints
import sys
from lit import Lit
from formula import ExpTable

# x_1 + ... + x_n <= 1
# ovar <=> x_1 \/ ... \/ x_n
def atmost_one(problem, elts, ovar = None):
  svars = [ elts[0] ] + [ problem.exist() for v in xrange(1, len(elts)-1) ]
  #print >> sys.stderr,"elts", elts
  #print >> sys.stderr, "problem at most svar", svars
  if ovar is not None:
    svars.append(ovar)

  for ei in xrange(1, len(elts)):
    problem.addClause([~svars[ei-1], ~elts[ei]])
  for ei in xrange(1, len(svars)):
    problem.addClause([~svars[ei-1], svars[ei]])
  for ei in xrange(1, len(svars)):
    problem.addClause([~elts[ei], svars[ei]])

  if ovar is not None:
    for ei in xrange(1, len(svars)):
      problem.addClause([~svars[ei], svars[ei-1], elts[ei]])
  return svars

def full_cmp(solver, a, b, c, d):
  # (c <-> a \/ b)
  solver.addClause([~c, a, b])
  solver.addClause([c, ~a])
  solver.addClause([c, ~b])

  # (d <-> a /\ b)
  solver.addClause([d, ~a, ~b])
  solver.addClause([~d, a])
  solver.addClause([~d, b])

def half_cmp(solver, a, b, c, d):
  # (c <-> a \/ b)
  solver.addClause([c, ~a])
  solver.addClause([c, ~b])

  # (d <-> a /\ b)
  solver.addClause([d, ~a, ~b])

## Construct a direct sorter with clauses:
## /\ S -> y_{|S|}
# the above function in one statement
def powerset(lst):
  return reduce(lambda result, x:
          result + [subset + [x] for subset in result], lst, [[]])

def choices(s, k):
  if len(s) <= k:
    return [s]
  return choices(s[1:], k) + map(lambda elts: [s[0]] + elts, choices(s[1:], k-1))

def direct_sort(problem, xs):
  os = [ problem.exist() for x in xs ]
  for s in powerset(xs):
    if len(s) == 0:
      continue
    problem.addClause([~os[len(s)-1]] + map(lambda x: ~x, s))
  return os

def direct_smerge(problem, xs, ys, k):
  assert len(xs) + len(ys) >= k

  os = [ problem.exist() for xi in xrange(k) ]
  for m in xrange(min(len(xs), k)):
    problem.addClause([~xs[m], os[m]])
  for m in xrange(min(len(ys), k)):
    problem.addClause([~ys[m], os[m]])
  for mx in xrange(min(len(xs), k-1)):
    for my in xrange(min(len(ys), k-mx-1)):
      problem.addClause([~xs[mx], ~ys[my], os[mx+my+1]])
  ## GKG - Check
  for v in xrange(k):
    ## sum(xs, ys) >= v+1
    for m in xrange(max(0, v-len(ys)), min(v-1, len(xs))):
      problem.addClause([xs[m], ys[v-m], ~os[v]])

  return os

def direct_card(problem, xs, k):
  os = [ problem.exist() for xi in xrange(k) ]
  for m in xrange(k):
    for xsub in choices(xs, m+1):
      problem.addClause([os[m]] + map(lambda x: ~x, xsub))
  return os

def oddeven_merge(problem, xs, ys, comp):
  lx = len(xs)
  ly = len(ys)

  if lx == 0:
    return ys
  if ly == 0:
    return xs
  if lx == 1 and ly == 1:
    out = [problem.exist(), problem.exist()]
#    print xs, ys, out
    comp(problem, xs[0], ys[0], out[0], out[1])
    return out

  xeven = [xs[i] for i in xrange(0, len(xs), 2)]
  xodd = [xs[i] for i in xrange(1, len(xs), 2)]
  yeven = [ys[i] for i in xrange(0, len(ys), 2)]
  yodd = [ys[i] for i in xrange(1, len(ys), 2)]

  zeven = oddeven_merge(problem, xeven, yeven, comp)
  zodd = oddeven_merge(problem, xodd, yodd, comp)

  out = [zeven[0]]
  for i in xrange(1, len(zeven)):
    [zlow, zhigh] = [problem.exist(), problem.exist()]
    comp(problem, zeven[i], zodd[i-1], zlow, zhigh)
    out.append(zlow)
    out.append(zhigh)
  if lx%2 == ly%2:
    out.append(zodd[-1])
  return out

def oddeven_sort(problem, elts, comp):
  if len(elts) <= 1:
    return elts

  # elts is of even length.
  bs = [ elts[vi] for vi in xrange(0, len(elts), 2) ]
  cs = [ elts[vi] for vi in xrange(1, len(elts), 2) ]

  bsort = oddeven_sort(problem, bs, comp)
  csort = oddeven_sort(problem, cs, comp)

  return oddeven_merge(problem, bsort, csort, comp);

## Construct a sorting network.
def sorter(problem, elts, comp = full_cmp):
  return oddeven_sort(problem, elts, comp)

def smerge(problem, xs, ys, vmax, comp):
  xs = xs[:vmax]
  ys = ys[:vmax]

  if len(xs) + len(ys) == 0:
    return []

  if len(xs) == 0:
    return ys
  if len(ys) == 0:
    return xs

  if vmax == 1:
    z = problem.exist()
    problem.addClause([~xs[0], z])
    problem.addClause([~ys[0], z])
    ## GKG: Added
    problem.addClause([~z, xs[0], ys[0]])
    return [z]
  
  if len(xs) + len(ys) <= vmax:
    return oddeven_merge(problem, xs, ys, comp)
  
  xeven = [xs[i] for i in xrange(0, len(xs), 2)]
  xodd = [xs[i] for i in xrange(1, len(xs), 2)]
  yeven = [ys[i] for i in xrange(0, len(ys), 2)]
  yodd = [ys[i] for i in xrange(1, len(ys), 2)]

  if vmax%2 == 0:
    meven = smerge(problem, xeven, yeven, (vmax/2)+1, comp)
    modd = smerge(problem, xodd, yodd, vmax/2, comp)
  
#    print xs, ys, vmax, "~~>", (xeven, xodd), (yeven, yodd), "~~>", meven, modd
    out = [meven[0]]
    for ii in xrange(1, len(meven)-1):
      [zlow, zhigh] = [problem.exist(), problem.exist()]
      comp(problem, meven[ii], modd[ii-1], zlow, zhigh)
      out.append(zlow)
      out.append(zhigh)
    zmax = problem.exist()
    problem.addClause([~meven[-1], zmax])
    problem.addClause([~modd[-1], zmax])
    # GKG: Added
    problem.addClause([~zmax, meven[-1], modd[-1]])

    out.append(zmax)
    return out
  else:
    meven = smerge(problem, xeven, yeven, (vmax+1)/2, comp)
    modd = smerge(problem, xodd, yodd, (vmax-1)/2, comp)
  
    out = [meven[0]]
    for ii in xrange(1, len(meven)):
      [zlow, zhigh] = [problem.exist(), problem.exist()]
      comp(problem, meven[ii], modd[ii-1], zlow, zhigh)
      out.append(zlow)
      out.append(zhigh)
    return out

def cardnet(problem, elts, vmax, comp = full_cmp):
#  return sorter(problem, elts, comp)[:vmax]

  ## Other version
  if len(elts) <= vmax:
    return sorter(problem, elts, comp)
  
  ## Figure out some split.
  split = len(elts)/2
#  split = vmax
  xs = cardnet(problem, elts[:split], vmax, comp)
  ys = cardnet(problem, elts[split:], vmax, comp)
  return smerge(problem, xs, ys, vmax, comp)
#  return direct_smerge(problem, xs, ys, vmax)

## Currently a naive tseitin transformation
def reify(problem, formula, cache = {}):
  if formula in cache:
    return cache[formula]

  kind = formula.kind()
  if kind == ExpTable.LEAF:
    child = formula.children()[0]
    if child == ExpTable.T:
      ret = problem.ttt()
    elif child == ExpTable.F:
      ret = problem.fff()
    else:
      ret = Lit(child)
  elif kind == ExpTable.NOT:
    cv = reify(problem, formula.children()[0], cache)
    return ~cv
  else:
    rv = problem.exist()
    [c0, c1] = formula.children()
    v0 = reify(problem, c0, cache)
    v1 = reify(problem, c1, cache)
#    print >> sys.stderr, rv, "<=>", v0, "op", v1
    if kind == ExpTable.AND:
      problem.addClause([v0, ~rv])
      problem.addClause([v1, ~rv])
      problem.addClause([~v0, ~v1, rv])
    else:
      ## kind == ExpTable.OR
      problem.addClause([~v0, rv])
      problem.addClause([~v1, rv])
      problem.addClause([v0, v1, ~rv])
    ret = rv

  cache[formula] = ret
  return ret
